/*
                  Endeavour Mark II - VFS Browser Callbacks
 */

#ifndef BROWSER_CB_H
#define BROWSER_CB_H

#include <gtk/gtk.h>
#include "edv_device.h"
#include "edv_mime_type.h"
#include "browser.h"


/* GTK+ Signal Callbacks */
extern void EDVBrowserRealizeCB(GtkWidget *widget, gpointer data);
extern gint EDVBrowserDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
extern gint EDVBrowserKeyEventCB(
	 GtkWidget *widget, GdkEventKey *key, gpointer data
);
extern gint EDVBrowserButtonPressEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

extern void EDVBrowserHandleChildAttachedCB(
	GtkHandleBox *handle_box, GtkWidget *child, gpointer data
);
extern void EDVBrowserHandleChildDetachedCB(
	GtkHandleBox *handle_box, GtkWidget *child, gpointer data
);

extern void EDVBrowserLocBarIconRealizeCB(GtkWidget *widget, gpointer data);

extern void EDVBrowserTreeSelectRowCB(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
);
extern void EDVBrowserTreeUnselectRowCB(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
);
extern void EDVBrowserTreeExpandCB(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
);
extern void EDVBrowserTreeCollapseCB(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
);

extern void EDVBrowserListRealizeCB(GtkWidget *widget, gpointer data);
extern void EDVBrowserResizeColumnCB(
	GtkCList *clist, gint column, gint width, gpointer data
);
extern void EDVBrowserClickColumnCB(
	GtkCList *clist, gint column, gpointer data
);
extern void EDVBrowserSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
extern void EDVBrowserUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);

extern void EDVBrowserLocationComboActivateCB(GtkWidget *widget, gpointer data);

/* Menu Item Callbacks */
extern void EDVBrowserMenuItemCB(GtkWidget *widget, gpointer data);
extern gint EDVBrowserMenuItemEnterCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
extern gint EDVBrowserMenuItemLeaveCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
extern void EDVBrowserMenuItemCmdObjectOpCB(GtkWidget *widget, gpointer data);
extern void EDVBrowserMenuItemCmdNewObjectCB(GtkWidget *widget, gpointer data);
extern gint EDVBrowserMenuItemCmdEnterCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
extern gint EDVBrowserMenuItemCmdLeaveCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);

/* Paths List Get Callback */
extern GList *EDVBrowserGetSelectedPathsListCB(gpointer data);

/* Go To Directory Callback */
extern void EDVBrowserGotoDirectoryCB(gpointer data, const gchar *path);

/* Mount Bar Callbacks */
extern void EDVBrowserMountBarMountCB(
	GtkWidget *widget, const gint dev_num, EDVDevice *dev,
	gpointer data
);
extern void EDVBrowserMountBarEjectCB(
	GtkWidget *widget, const gint dev_num, EDVDevice *dev,
	gpointer data
);
extern void EDVBrowserMountBarGoToCB(
	GtkWidget *widget, const gint dev_num, EDVDevice *dev,
	gpointer data
);

/* Find Bar Callbacks */
extern const gchar *EDVBrowserFindBarLocationCB(
	GtkWidget *bar,
	gpointer data
);
extern void EDVBrowserFindBarStartCB(
	GtkWidget *bar,
	gpointer data
);
extern void EDVBrowserFindBarEndCB(
	GtkWidget *bar,
	const gint nmatches,
	gpointer data
);
extern void EDVBrowserFindBarMatchCB(
	GtkWidget *bar,
	const gchar *path,
	GList *properties_list,
	const gint line_index,
	const gchar *excerpt,
	gpointer data
);

/* Status Bar Callbacks */
extern void EDVBrowserStatusMessageCB(
	GtkWidget *widget,
	const gchar *message,
	gpointer data
);
extern void EDVBrowserStatusProgressCB(
	GtkWidget *widget,
	const gfloat progress,
	gpointer data
);

/* Window Created/Deleted Callbacks */
extern void EDVBrowserWindowCreatedCB(
	edv_browser_struct *browser,
	const EDVWindowType win_type,
	gpointer win
);
extern void EDVBrowserWindowDeletedCB(
	edv_browser_struct *browser,
	const EDVWindowType win_type,
	gpointer win
);

/* Master Write Protect Changed Callback */
extern void EDVBrowserWriteProtectChangedCB(
	edv_browser_struct *browser,
	const gboolean state
);

/* Delete Method Changed Callback */
extern void EDVBrowserDeleteMethodChangedCB(
	edv_browser_struct *browser,
	const EDVDeleteMethod delete_method
);

/* Object Callbacks */
extern void EDVBrowserObjectAddedCB(
	edv_browser_struct *browser,
	const gchar *path,
	EDVVFSObject *obj
);
extern void EDVBrowserObjectModifiedCB(
	edv_browser_struct *browser,
	const gchar *path,
	const gchar *new_path,
	EDVVFSObject *obj
);
extern void EDVBrowserObjectRemovedCB(
	edv_browser_struct *browser,
	const gchar *path
);

/* Mount/Unmount Callback */
extern void EDVBrowserMountCB(
	edv_browser_struct *browser,
	const gint dev_num, EDVDevice *d,
	const gboolean mounted
);

/* Recycled Object Callbacks */
extern void EDVBrowserRecycledObjectAddedCB(
	edv_browser_struct *browser,
	const guint index
);
extern void EDVBrowserRecycledObjectModifiedCB(
	edv_browser_struct *browser,
	const guint index
);
extern void EDVBrowserRecycledObjectRemovedCB(
	edv_browser_struct *browser,
	const guint index
);

/* Reconfigured Callback */
extern void EDVBrowserReconfiguredCB(edv_browser_struct *browser);

/* MIME Type Callbacks */
extern void EDVBrowserMimeTypeAddedCB(
	edv_browser_struct *browser,
	const gint mt_num, EDVMIMEType *m
);
extern void EDVBrowserMimeTypeModifiedCB(
	edv_browser_struct *browser,
	const gint mt_num, EDVMIMEType *m
);
extern void EDVBrowserMimeTypeRemovedCB(
	edv_browser_struct *browser,
	const gint mt_num
);

/* Device Callbacks */
extern void EDVBrowserDeviceAddedCB(
	edv_browser_struct *browser,
	const gint dev_num, EDVDevice *d
);
extern void EDVBrowserDeviceModifiedCB(
	edv_browser_struct *browser,
	const gint dev_num, EDVDevice *d
);
extern void EDVBrowserDeviceRemovedCB(
	edv_browser_struct *browser,
	const gint dev_num
);


#endif	/* BROWSER_CB_H */
