/*
			    Endeavour Operations
 */

#ifndef EDV_OP_H
#define EDV_OP_H

#include <gtk/gtk.h>

#include "url.h"
#include "edv_types.h"
#include "libendeavour2-base/edv_vfs_obj.h"
#include "libendeavour2-base/edv_recycled_obj.h"
#include "libendeavour2-base/edv_archive_obj.h"
#include "obj_op_dlg.h"
#include "browser.h"
#include "imbr.h"
#include "archiver.h"
#include "recbin.h"
#include "prop_dlg.h"


/* New/Map Window Nexus */
extern gint EDVNewWindow(
	EDVCore *core,
	const EDVWindowType win_type,
	const gchar *path,
	const gchar *extra,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);
extern gint EDVNewWindowName(
	EDVCore *core,
	const gchar *win_name,
	const gchar *path,
	const gchar *extra,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);


/* New/Delete/Map/Unmap Windows
 *
 * About Dialog
 */
extern void EDVMapAboutDlgPage(
	EDVCore *core,
	const gchar *page_name,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);
extern void EDVMapAboutDlg(
	EDVCore *core,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);

/* File Browser */
extern gint EDVNewBrowser(
	EDVCore *core,
	const gchar *path,
	const gchar *extra,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry
);
extern void EDVDeleteBrowser(
	EDVCore *core,
	const gint i
);

/* Image Browser */
extern gint EDVNewImbr(
	EDVCore *core,
	const gchar *path,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry
);
extern void EDVDeleteImbr(
	EDVCore *core,
	const gint i
);

/* Archiver */
extern gint EDVNewArchiver(
	EDVCore *core,
	const gchar *path,
	const gchar *password,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry
);
extern void EDVDeleteArchiver(
	EDVCore *core,
	const gint i
);

/* Recycle Bin & Recycle Bin Desktop Icon */
extern void EDVMapRecBinDeskIcon(EDVCore *core);
extern void EDVMapRecBin(
	EDVCore *core,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry
);

/* MIME Types List Window */
extern void EDVMapMIMETypesListWin(
	EDVCore *core,
	const gchar *type,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);

/* Devices List Window */
extern void EDVMapDevicesListWin(
	EDVCore *core,
	const gchar *path,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);

/* History List Window */
extern void EDVMapHistoryListWin(
	EDVCore *core,
	const gint event_index,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);

/* Options Window */
extern void EDVMapOptionsWinPage(
	EDVCore *core,
	const gchar *page_name,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);
extern void EDVMapOptionsWin(
	EDVCore *core,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);

/* Customize Window */
extern void EDVMapCustomizeWinPage(
	EDVCore *core,
	const gchar *page_name,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);
extern void EDVMapCustomizeWin(
	EDVCore *core,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);

/* Properties Dialog */
extern gint EDVNewPropertiesDialogVFS(
	EDVCore *core,
	const gchar *path,
	const gchar *page_name,
	GList *extended_properties_list,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);
extern gint EDVNewPropertiesDialogRecycled(
	EDVCore *core,
	const gulong index,
	const gchar *page_name,
	GList *extended_properties_list,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);
extern gint EDVNewPropertiesDialogArchive(
	EDVCore *core,
	const gchar *path,
	const gchar *arch_path,
	const gchar *page_name,
	GList *extended_properties_list,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry,
	GtkWidget *toplevel
);


/* Running */
extern void EDVRunDeviceCheck(
	EDVCore *core,
	EDVDevice *d,
	GtkWidget *toplevel
);
extern void EDVRunDeviceTools(
	EDVCore *core,
	EDVDevice *d,
	GtkWidget *toplevel
);
extern void EDVRunDeviceFormat(
	EDVCore *core,
	EDVDevice *d,
	GtkWidget *toplevel
);

extern void EDVRunTerminal(
	EDVCore *core,
	const gchar *cmd,
	const gchar *wd,
	GtkWidget *toplevel
);


/* Object Operations Dialog */
extern edv_obj_op_dlg_struct *EDVGetObjOpDlg(EDVCore *core);


/* Window Mapping */
extern void EDVMapRunDialogCommand(
	EDVCore *core,
	const gchar *command,
	const gchar *working_dir,
	GtkWidget *toplevel
);
extern void EDVMapRunDialog(
	EDVCore *core,
	GtkWidget *toplevel
);

extern void EDVMapFindWin(
	EDVCore *core,
	const EDVLocationType location_type,
	const gchar *location,
	const gchar *value
);
extern void EDVMapBrowserFindWin(
	EDVCore *core,
	edv_browser_struct *browser
);
extern void EDVMapImbrFindWin(
	EDVCore *core,
	edv_imbr_struct *imbr
);
extern void EDVMapArchiverFindWin(
	EDVCore *core,
	EDVArchiver *archiver
);
extern void EDVMapRecBinFindWin(
	EDVCore *core,
	edv_recbin_struct *recbin
);


/* Update Resources & Widgets */
extern void EDVUpdateRCStyles(EDVCore *core);
extern void EDVUpdateIDPULists(EDVCore *core);
extern void EDVUpdateDevicesPUList(EDVCore *core);
extern void EDVUpdateMIMETypeHintIndices(EDVCore *core);
extern void EDVUpdateOpenWithPUList(EDVCore *core);


/* Refresh & Reset */
extern void EDVRefresh(EDVCore *core);
extern void EDVReset(EDVCore *core);


/* Sync */
extern void EDVSyncEDV(EDVCore *core);


/* Clearing */
extern void EDVClearErrorMessages(EDVCore *core);
extern void EDVClearAllHistory(
	EDVCore *core,
	const gboolean confirm,
	GtkWidget *toplevel
);
extern void EDVClearEventsHistory(
	EDVCore *core,
	const gboolean confirm,
	GtkWidget *toplevel
);
extern void EDVClearLocationBarsHistory(
	EDVCore *core,
	const gboolean confirm,
	GtkWidget *toplevel
);
extern void EDVClearRunHistory(
	EDVCore *core,
	const gboolean confirm,
	GtkWidget *toplevel
);

/* Recycle Bin Size Check */
extern void EDVRecycleBinSizeCheck(
	EDVCore *core,
	GtkWidget *toplevel
);

/* Purge Recycle Bin */
extern void EDVPurgeAllRecycledObjects(
	EDVCore *core,
	const gboolean map_recbin,
	const gboolean show_progress,
	const gboolean interactive,
	GtkWidget *toplevel
);

/* Query Unmount Before Exit */
extern void EDVQueryUnmountBeforeExit(EDVCore *core);

/* Network */
extern gint EDVInternetDownloadObject(
	EDVCore *core,
	const URLStruct *url,
	const gchar *target_path,
	GtkWidget *toplevel
);


#endif	/* EDV_OP_H */
