/*
	       Endeavour Mark II - Utilities (GTK+ Dependent)
 */

#ifndef EDV_UTILS_GTK_H
#define EDV_UTILS_GTK_H

#include <gtk/gtk.h>

#include "cfg.h"
#include "edv_types.h"
#include "edv_pixmap.h"
#include "edv_cursor.h"
#include "edv_device.h"
#include "edv_mime_type.h"
#include "libendeavour2-base/edv_vfs_obj.h"


/* Accelerator Key Matching */
extern gint EDVMatchAccelKeyOPID(
	const CfgList *cfg_list,
	const gchar *parm,
	const guint accel_key, const guint accel_mods
);
extern CfgAccelkey *EDVMatchAccelKey(
	const CfgList *cfg_list,
	const gchar *parm,
	const gint opid
);

/* EDVPixmaps */
extern EDVPixmap *EDVLoadPixmapFromData(
	EDVCore *core,
	edv_pixmap_data *data,
	const gchar *name
);
extern EDVPixmap *EDVOpenPixmapFromFile(
	EDVCore *core,
	const gchar *path
);
extern EDVPixmap *EDVNewProgressPixmap(
	GdkWindow *window,
	GtkStyle *style,
	const gint width, const gint height,
	const gfloat coeff,
	const gboolean draw_value,
	const GtkOrientation orientation,
	const gboolean reverse
);
extern GtkWidget *EDVNewPixmapWidget(
	EDVCore *core,
	edv_pixmap_data *data,
	const gchar *name
);
extern gchar *EDVGetRecycleBinIconPath(
	EDVCore *core,
	const EDVIconSize size,
	const gint nrecycled_objects
);
extern edv_pixmap_data *EDVGetRecycleBinIconData(
	const EDVIconSize size,
	const gint nrecycled_objects,
	const gchar **c_declaration_id_rtn
);

/* Cursors */
extern GdkCursor *EDVGetCursor(
	EDVCore *core,
	const edv_cursor_code code
);

/* GtkCTree Nodes */
extern gint EDVNodeGetLevel(GtkCTreeNode *node);
extern GtkCTreeNode *EDVNodeGetParent(GtkCTreeNode *node);
extern GtkCTreeNode *EDVNodeGetChild(GtkCTreeNode *node);
extern GtkCTreeNode *EDVNodeGetSibling(GtkCTreeNode *node);
extern GtkCTreeNode *EDVNodeGetToplevel(GtkCTree *ctree);

/* GtkCList & GtkCTree Utilities */
extern gint EDVCListGetSelectedLast(
	GtkCList *clist,
	GtkCListRow **row_ptr_rtn
);
extern GtkCTreeNode *EDVCTreeGetSelectedLast(
	GtkCTree *ctree,
	GtkCTreeRow **row_ptr_rtn
);
extern gboolean EDVNodeGetIndex(
	GtkCTree *ctree,
	GtkCTreeNode *node,
	gint *row_rtn, gint *column_rtn
);
extern GtkCTreeNode *EDVNodeGetByCoordinates(
	GtkCTree *ctree,
	const gint x, const gint y
);
extern void EDVScrollCListToPosition(
	GtkCList *clist,
	const gfloat hpos, const gfloat vpos
);

/* Endeavour Window Matching */
extern gboolean EDVMatchWindowFromToplevel(
	EDVCore *core, GtkWidget *toplevel,
	gint *browser_num,
	gint *imbr_num,
	gint *archiver_num,
	gint *recbin_num
);

/* Object Checking */
extern gboolean EDVIsObjectNameValid(const gchar *name);
extern gboolean EDVIsObjectHidden(EDVVFSObject *obj);

/* Permission Checking */
extern gboolean EDVCheckUIDIsOwner(
	const gint effective_uid,
	const gint owner_id
);
extern gboolean EDVCheckIsOwner(
	EDVCore *core,
	const gint owner_id
);
extern gboolean EDVIsObjectAccessable(
	EDVCore *core,
	EDVVFSObject *obj
);
extern gboolean EDVIsObjectReadable(
	EDVCore *core,
	EDVVFSObject *obj
);
extern gboolean EDVIsObjectWriteable(
	EDVCore *core,
	EDVVFSObject *obj
);
extern gboolean EDVIsObjectExecutable(
	EDVCore *core,
	EDVVFSObject *obj
);

/* Internal Supported Object Type Checking */
extern gboolean EDVCheckEDVArchiverArchive(
	EDVCore *core,
	const gchar *name
);

/* Message Output */
extern void edv_message(
	const gchar *title,
	const gchar *message,
	const gchar *details,
	GtkWidget *toplevel
);
extern void edv_message_info(
	const gchar *title,
	const gchar *message,
	const gchar *details,
	GtkWidget *toplevel
);
extern void edv_message_warning(
	const gchar *title,
	const gchar *message,
	const gchar *details,
	GtkWidget *toplevel
);
extern void edv_message_error(
	const gchar *title,
	const gchar *message,
	const gchar *details,
	GtkWidget *toplevel
);
extern void EDVMessageObjectOPError(
	const gchar *title,
	const gchar *message,
	const gchar *path,
	GtkWidget *toplevel
);
extern void edv_message_from_file(
	const gchar *path,
	const gchar *title,
	gint cdialog_icon,
	GtkWidget *toplevel
);

/* Play Sound */
extern void edv_beep(EDVCore *core);
extern void edv_play_sound_beep(EDVCore *core);
extern void edv_play_sound_info(EDVCore *core);
extern void edv_play_sound_question(EDVCore *core);
extern void edv_play_sound_warning(EDVCore *core);
extern void edv_play_sound_error(EDVCore *core);
extern void edv_play_sound_completed(EDVCore *core);

/* Menu & Button Interaction */
extern void EDVMapMenuToButton(GtkWidget *menu, GtkWidget *button);

/* Window Centering */
extern void EDVCenterWindowToWindow(GtkWidget *w1, GtkWidget *w2);
extern void EDVCenterWindowToWindowOffset(
	GtkWidget *w1, GtkWidget *w2, gint x_offset, gint y_offset
);

/* GtkEditable & GtkEntry Endowments */
extern void EDVEntrySetDND(EDVCore *core, GtkWidget *w);
extern void EDVEntrySetCompletePath(EDVCore *core, GtkWidget *w);

/* Scrolled Window Setup Utilities */
extern GtkWidget *EDVScrolledWindowNew(
	GtkPolicyType hscrollbar_policy,
	GtkPolicyType vscrollbar_policy,
	GtkWidget **event_box_rtn,
	GtkWidget **vbox_rtn
);

/* Copy Object Path Utilities */
extern void EDVCopyDiskObjectsToDDEPath(
	EDVCore *core,
	GList *objs_list,
	GtkWidget *owner
);
extern void EDVCopyDiskObjectsToDDEURL(
	EDVCore *core,
	GList *objs_list,
	GtkWidget *owner
);

/* Style Utilities */
extern GtkRcStyle *EDVCreateRCStyleFromCfg(
	const CfgStyle *style
);

/* History Utilities */
extern void EDVAppendHistory(
	EDVCore *core,
	const EDVHistoryType type,
	const gulong time_start,
	const gulong time_end,
	const gint status,
	const gchar *src_path,
	const gchar *tar_path,
	const gchar *comments
);

/* Write Protect Utilities */
extern gboolean EDVCheckWriteProtect(
	EDVCore *core,
	const gboolean verbose,
	GtkWidget *toplevel
);

/* Query Name Filter */
extern gboolean EDVQueryNameFilter(
	EDVCore *core,
	gchar **filter_rtn,
	GtkWidget *toplevel
);


#endif	/* EDV_UTILS_GTK_H */
