/*
                Endeavour Mark II - Image Browser Callbacks
 */

#ifndef IMBR_CB_H
#define IMBR_CB_H

#include <gtk/gtk.h>
#include "edv_device.h"
#include "edv_mime_type.h"
#include "imbr.h"


/* GTK+ Signal Callbacks */
extern void EDVImbrRealizeCB(GtkWidget *widget, gpointer data);
extern gint EDVImbrDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
extern gint EDVImbrKeyEventCB(
	 GtkWidget *widget, GdkEventKey *key, gpointer data
);
extern gint EDVImbrButtonPressEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

extern void EDVImbrHandleChildAttachedCB(
	GtkHandleBox *handle_box, GtkWidget *child, gpointer data
);
extern void EDVImbrHandleChildDetachedCB(
	GtkHandleBox *handle_box, GtkWidget *child, gpointer data
);

extern void EDVImbrLocBarIconRealizeCB(GtkWidget *widget, gpointer data);

extern void EDVImbrTListRealizeCB(GtkWidget *widget, gpointer data);
extern void EDVImbrTListSelectCB(
	tlist_struct *tlist, GdkEventButton *button, gint thumb_num,
	gpointer data
);
extern void EDVImbrTListUnselectCB(
	tlist_struct *tlist, GdkEventButton *button, gint thumb_num,
	gpointer data
);

extern void EDVImbrLocationComboActivateCB(GtkWidget *widget, gpointer data);

extern gint EDVImbrLoadThumbsPassiveTOCB(gpointer data);

/* Menu Item Callbacks */
extern void EDVImbrMenuItemCB(GtkWidget *widget, gpointer data);
extern gint EDVImbrMenuItemEnterCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
extern gint EDVImbrMenuItemLeaveCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
extern void EDVImbrMenuItemCmdObjectOpCB(GtkWidget *widget, gpointer data);
extern void EDVImbrMenuItemCmdNewObjectCB(GtkWidget *widget, gpointer data);
extern gint EDVImbrMenuItemCmdEnterCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
extern gint EDVImbrMenuItemCmdLeaveCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);

/* Paths List Get Callback */
extern GList *EDVImbrGetSelectedPathsListCB(gpointer data);

/* Go To Directory Callback */
extern void EDVImbrGotoDirectoryCB(gpointer data, const gchar *path);

/* Mount Bar Callbacks */
extern void EDVImbrMountBarMountCB(
	GtkWidget *widget, const gint dev_num, EDVDevice *dev,
	gpointer data
);
extern void EDVImbrMountBarEjectCB(
	GtkWidget *widget, const gint dev_num, EDVDevice *dev,
	gpointer data
);
extern void EDVImbrMountBarGoToCB(
	GtkWidget *widget, const gint dev_num, EDVDevice *dev,
	gpointer data
);

/* Find Bar Callbacks */
extern const gchar *EDVImbrFindBarLocationCB(
	GtkWidget *bar,
	gpointer data
);
extern void EDVImbrFindBarStartCB(
	GtkWidget *bar,
	gpointer data
);
extern void EDVImbrFindBarEndCB(
	GtkWidget *bar,
	const gint nmatches,
	gpointer data
);
extern void EDVImbrFindBarMatchCB(
	GtkWidget *bar,
	const gchar *path,
	GList *properties_list,
	const gint line_index,
	const gchar *excerpt,
	gpointer data
);

/* Status Bar Callbacks */
extern void EDVImbrStatusMessageCB(
	GtkWidget *widget,
	const gchar *message,
	gpointer data
);
extern void EDVImbrStatusProgressCB(
	GtkWidget *widget,
	const gfloat progress,
	gpointer data
);

/* Master Write Protect Changed Callback */
extern void EDVImbrWriteProtectChangedCB(
	edv_imbr_struct *imbr,
	const gboolean state
);

/* Delete Method Changed Callback */
extern void EDVImbrDeleteMethodChangedCB(
	edv_imbr_struct *imbr,
	const EDVDeleteMethod delete_method
);

/* Object Callbacks */
extern void EDVImbrObjectAddedCB(
	edv_imbr_struct *imbr,
	const gchar *path,
	EDVVFSObject *obj
);
extern void EDVImbrObjectModifiedCB(
	edv_imbr_struct *imbr,
	const gchar *path,
	const gchar *new_path,
	EDVVFSObject *obj
);
extern void EDVImbrObjectRemovedCB(
	edv_imbr_struct *imbr,
	const gchar *path
);

/* Mount/Unmount Callback */
extern void EDVImbrMountCB(
	edv_imbr_struct *imbr,
	const gint dev_num, EDVDevice *d,
	const gboolean mounted
);

/* Recycled Object Callbacks */
extern void EDVImbrRecycledObjectAddedCB(
	edv_imbr_struct *imbr,
	const guint index
);
extern void EDVImbrRecycledObjectModifiedCB(
	edv_imbr_struct *imbr,
	const guint index
);
extern void EDVImbrRecycledObjectRemovedCB(
	edv_imbr_struct *imbr,
	const guint index
);

/* Window Created/Deleted Callbacks */
extern void EDVImbrWindowCreatedCB(
	edv_imbr_struct *imbr,
	const EDVWindowType win_type,
	gpointer win
);
extern void EDVImbrWindowDeletedCB(
	edv_imbr_struct *imbr, 
	const EDVWindowType win_type,
	gpointer win
);

/* Reconfigured Callback */
extern void EDVImbrReconfiguredCB(edv_imbr_struct *imbr);

/* MIME Type Callbacks */
extern void EDVImbrMimeTypeAddedCB(
	edv_imbr_struct *imbr,
	const gint mt_num, EDVMIMEType *m
);
extern void EDVImbrMimeTypeModifiedCB(
	edv_imbr_struct *imbr,
	const gint mt_num, EDVMIMEType *m
);
extern void EDVImbrMimeTypeRemovedCB(
	edv_imbr_struct *imbr,
	const gint mt_num
);

/* Device Callbacks */
extern void EDVImbrDeviceAddedCB(
	edv_imbr_struct *imbr,
	const gint dev_num, EDVDevice *d
);
extern void EDVImbrDeviceModifiedCB(
	edv_imbr_struct *imbr,
	const gint dev_num, EDVDevice *d
);
extern void EDVImbrDeviceRemovedCB(
	edv_imbr_struct *imbr,
	const gint dev_num
);


#endif	/* IMBR_CB_H */
