/*
                   Endeavour Mark II - Properties Dialog
 */

#ifndef PROP_DLG_H
#define PROP_DLG_H

#include <gtk/gtk.h>

#include "edv_types.h"
#include "libendeavour2-base/edv_vfs_obj.h"
#include "edv_device.h"
#include "edv_mime_type.h"
#include "mime_type_edit_dlg.h"
#include "device_edit_dlg.h"


typedef struct _EDVPropertiesDialog	EDVPropertiesDialog;
#define EDV_PROPERTIES_DIALOG(p)	((EDVPropertiesDialog *)(p))


/*
 *	Properties Dialog:
 */
struct _EDVPropertiesDialog {

	GtkWidget       *toplevel;
	GtkAccelGroup   *accelgrp;
	gint            busy_count,
			freeze_count;
	gboolean	has_changes;
	gchar		*title;
	EDVCore *core;

	GtkWidget	*main_vbox,
			*notebook,
			*ok_btn,
			*apply_btn,
			*cancel_btn,
			*close_btn;

	/* Object's Properties */
	EDVLocationType	location_type;
	GList		*properties_list,	/* Basic object properties, a
						 * GList of EDVProperty *
						 * properties */
			*meta_data_list;	/* Format and type-specific
						 * properties, a GList of
						 * EDVProperty * properties */
	gchar		*mime_type,
			*arch_path;		/* Path to the archive that
						 * the object is in (if
						 * location_type is
						 * EDV_LOCATION_TYPE_ARCHIVE) */

	/* Pages List, GList of EDVPropertiesDialogPage * pages */
	GList		*pages_list;
	gint		cur_page_num;		/* Current displayed page or
						 * -1 for none */

	/* Icon Selection Dialog */
	GtkWidget	*icon_sel_dlg;

	/* MIME Type Edit Dialog */
	edv_mime_type_edit_dlg_struct	*mime_type_edit_dlg;

	/* Device Edit Dialog */
	edv_device_edit_dlg_struct	*device_edit_dlg;
};


/* Callbacks */
extern void EDVPropertiesDialogReconfiguredCB(EDVPropertiesDialog *d);
extern void EDVPropertiesDialogWriteProtectChangedCB(
	EDVPropertiesDialog *d,
	const gboolean state
);
extern void EDVPropertiesDialogDeleteMethodChangedCB(
	EDVPropertiesDialog *d,
	const EDVDeleteMethod delete_method
);

extern void EDVPropertiesDialogMIMETypeAddedCB(
	EDVPropertiesDialog *d,
	const gint mt_num, EDVMIMEType *m
);
extern void EDVPropertiesDialogMIMETypeModifiedCB(
	EDVPropertiesDialog *d,
	const gint mt_num, EDVMIMEType *m
);
extern void EDVPropertiesDialogMIMETypeRemovedCB(
	EDVPropertiesDialog *d,
	const gint mt_num
);

extern void EDVPropertiesDialogDeviceAddedCB(
	EDVPropertiesDialog *d,
	const gint dev_num, EDVDevice *dev
);
extern void EDVPropertiesDialogDeviceModifiedCB(
	EDVPropertiesDialog *d,
	const gint dev_num, EDVDevice *dev
);
extern void EDVPropertiesDialogDeviceRemovedCB(
	EDVPropertiesDialog *d,
	const gint dev_num
);

extern void EDVPropertiesDialogObjectAddedCB(
	EDVPropertiesDialog *d,
	const gchar *path,
	EDVVFSObject *obj
);
extern void EDVPropertiesDialogObjectModifiedCB(
	EDVPropertiesDialog *d,
	const gchar *path,
	const gchar *new_path,
	EDVVFSObject *obj
);
extern void EDVPropertiesDialogObjectRemovedCB(
	EDVPropertiesDialog *d, const gchar *path
);

extern void EDVPropertiesDialogMountNotifyCB(
	EDVPropertiesDialog *d,
	const gint dev_num, EDVDevice *dev,
	const gboolean mounted
);

extern void EDVPropertiesDialogRecycledObjectAddedCB(
	EDVPropertiesDialog *d,
	const gulong index
);
extern void EDVPropertiesDialogRecycledObjectModifiedCB(
	EDVPropertiesDialog *d,
	const gulong index
);
extern void EDVPropertiesDialogRecycledObjectRemovedCB(
	EDVPropertiesDialog *d,
	const gulong index
);

/* Utilities */
extern GList *EDVPropertiesDialogUpdatePropertiesList(
	EDVCore *core,
	GList *properties_list,
	const EDVLocationType location_type,
	const gchar *path,
	const gulong index,
	const gchar *arch_path,
	GList **meta_data_list_rtn,
	gint *error_code_rtn
);
extern void EDVPropertiesDialogSetWidgetChangedSignal(
	EDVPropertiesDialog *d,
	GtkWidget *w
);

/* Pages */
extern void EDVPropertiesDialogUpdatePagesEmit(EDVPropertiesDialog *d);

/* Properties Dialog */
extern EDVPropertiesDialog *EDVPropertiesDialogNew(
	EDVCore *core,
	const EDVLocationType location_type,
	const gchar *path,
	const gulong index,
	const gchar *arch_path,
	GList *meta_data_list,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry
);
extern GtkWidget *EDVPropertiesDialogCreateIconSelector(EDVPropertiesDialog *d);
extern gboolean EDVPropertiesDialogSetPage(
	EDVPropertiesDialog *d,
	const gchar *page_name
);
extern void EDVPropertiesDialogSetHasChanges(
	EDVPropertiesDialog *d,
	const gboolean has_changes
);
extern void EDVPropertiesDialogUpdateMenus(EDVPropertiesDialog *d);
extern void EDVPropertiesDialogSetBusy(EDVPropertiesDialog *d, const gboolean busy);
extern gboolean EDVPropertiesDialogIsMapped(EDVPropertiesDialog *d);
extern void EDVPropertiesDialogMap(EDVPropertiesDialog *d);
extern void EDVPropertiesDialogUnmap(EDVPropertiesDialog *d);
extern void EDVPropertiesDialogDelete(EDVPropertiesDialog *d);


#endif	/* PROP_DLG_H */
