#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>				/* mknod() */
#include <sys/types.h>				/* mknod() */
#include <sys/stat.h>				/* mknod() */
#include <gtk/gtk.h>
#include <unistd.h>				/* mknod() */

#include "cfg.h"

#include "guiutils.h"

#include "edv_types.h"
#include "libendeavour2-base/edv_utils.h"
#include "libendeavour2-base/edv_property.h"
#include "libendeavour2-base/edv_vfs_obj.h"
#include "libendeavour2-base/edv_recycled_obj.h"
#include "libendeavour2-base/edv_archive_obj.h"
#include "edv_utils_gtk.h"
#include "prop_dlg.h"
#include "prop_dlg_device_node.h"

#include "config.h"


typedef struct _EDVPropertiesDeviceNodePage	EDVPropertiesDeviceNodePage;
#define EDV_PROPERTIES_DEVICE_NODE_PAGE(p)	((EDVPropertiesDeviceNodePage *)(p))


/* Check Support */
gboolean EDVPropertiesDeviceNodePageCheckSupportCB(
	EDVPropertiesDialog *d,
	const EDVObjectType type,
	const EDVLocationType location_type,
	GList *properties_list
);

/* Create */
gpointer EDVPropertiesDeviceNodePageCreateCB(
	EDVPropertiesDialog *d,
	GtkWidget *parent 
);

/* Update */
void EDVPropertiesDeviceNodePageUpdateCB(
	EDVPropertiesDialog *d,
	const EDVObjectType type,
	const EDVLocationType location_type,
	GList *properties_list,
	const int error_code,
	gpointer data
);

/* Apply */
gboolean EDVPropertiesDeviceNodePageApplyVFSCB(
	EDVPropertiesDialog *d,
	const EDVObjectType type,
	GList *properties_list,
	gpointer data
);
gboolean EDVPropertiesDeviceNodePageApplyRecBinCB(
	EDVPropertiesDialog *d,
	const EDVObjectType type,
	GList *properties_list,
	gpointer data
);

/* Destroy */
void EDVPropertiesDeviceNodePageDestroyCB(gpointer data);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (gint)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Device Node Page:
 */
struct _EDVPropertiesDeviceNodePage {
	GtkWidget	*toplevel;
	gint		freeze_count;
	EDVPropertiesDialog	*properties_dialog;
	GtkWidget	*major_entry,
			*minor_entry;
};


/*
 *	Check support callback.
 */
gboolean EDVPropertiesDeviceNodePageCheckSupportCB(
	EDVPropertiesDialog *d,
	const EDVObjectType type,
	const EDVLocationType location_type,
	GList *properties_list
)
{
	if((type != EDV_OBJECT_TYPE_DEVICE_BLOCK) &&
	   (type != EDV_OBJECT_TYPE_DEVICE_CHARACTER)
	)
	    return(FALSE);

	return(TRUE);
}

/*
 *	Create callback.
 */
gpointer EDVPropertiesDeviceNodePageCreateCB(
	EDVPropertiesDialog *d,
	GtkWidget *parent
)
{
	const gint	border_major = 5,
			border_minor = 2;
	GtkWidget	*w,
			*parent2, *parent3;
	EDVPropertiesDeviceNodePage *p = EDV_PROPERTIES_DEVICE_NODE_PAGE(g_malloc0(
	    sizeof(EDVPropertiesDeviceNodePage)
	));
	if(p == NULL)
	    return(NULL);

	p->properties_dialog = d;
	p->toplevel = parent;

	p->freeze_count++;

	/* Device numbers GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"El Artefacto Numeros"
#elif defined(PROG_LANGUAGE_FRENCH)
"Numro de composant"
#elif defined(PROG_LANGUAGE_GERMAN)
"Vorrichtung Zhlt"
#elif defined(PROG_LANGUAGE_ITALIAN)
"I Numeri Di Congegno"
#elif defined(PROG_LANGUAGE_DUTCH)
"Apparaat Nummert"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Artifcio Numera"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Innretning Numbers"
#else
"Device Numbers"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Major GtkHBox */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;
	/* Major GtkLabel */
	w = gtk_label_new(
#if defined(PROG_LANGUAGE_SPANISH)
"Mayor:"
#elif defined(PROG_LANGUAGE_FRENCH)
"Majeur:"
#elif defined(PROG_LANGUAGE_GERMAN)
"Major:"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Maggiore:"
#elif defined(PROG_LANGUAGE_DUTCH)
"Majoor:"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Importante:"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Major:"
#else
"Major:"
#endif
	);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Major GtkEntry */
	p->major_entry = w = gtk_entry_new();
	gtk_widget_set_usize(w, 80, -1);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	EDVPropertiesDialogSetWidgetChangedSignal(d, w);
	GUIEditableEndowPopupMenu(w, 0);
	gtk_widget_show(w);

	/* Minor GtkHBox */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;
	/* Minor GtkLabel */
	w = gtk_label_new(
#if defined(PROG_LANGUAGE_SPANISH)
"El Menor:"
#elif defined(PROG_LANGUAGE_FRENCH)
"Mineur:"
#elif defined(PROG_LANGUAGE_GERMAN)
"Minderjhriger:"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Minore:"
#elif defined(PROG_LANGUAGE_DUTCH)
"Minderjarige:"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Menor:"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Mindrerig:"
#else
"Minor:"
#endif
	);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Minor GtkEntry */
	p->minor_entry = w = gtk_entry_new();
	gtk_widget_set_usize(w, 80, -1);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	EDVPropertiesDialogSetWidgetChangedSignal(d, w);
	GUIEditableEndowPopupMenu(w, 0);
	gtk_widget_show(w);


	p->freeze_count--;

	return(p);
}


/*
 *	Update callback.
 */
void EDVPropertiesDeviceNodePageUpdateCB(
	EDVPropertiesDialog *d,
	const EDVObjectType type,
	const EDVLocationType location_type,
	GList *properties_list,
	const int error_code,
	gpointer data
)
{
	gint		major,
			minor;
	EDVPropertiesDeviceNodePage *p = EDV_PROPERTIES_DEVICE_NODE_PAGE(data);

	p->freeze_count++;

        /* If there are changes then get the current values from this
         * page's GtkWidgets, otherwise get the current values from
         * the properties list and set those values to this page's
         * GtkWidgets, this is to ensure that this page's GtkWidgets
         * get their values initially set for the first time this
         * update callback is called and to prevent subsequent
         * modification of the values displayed on the GtkWidgets   
         * after the user has changed the values on the GtkWidgets
         */
        if(d->has_changes)
        {
	    major = ATOI(gtk_entry_get_text(GTK_ENTRY(p->major_entry)));
	    minor = ATOI(gtk_entry_get_text(GTK_ENTRY(p->minor_entry)));
        }
	else
	{
	    gchar *s;

	    edv_device_numbers_parse(
		edv_properties_list_get_i(
		    properties_list,
		    "device_type"
		),
		&major, &minor
	    );

	    s = g_strdup_printf("%i", major);
	    gtk_entry_set_text(GTK_ENTRY(p->major_entry), s);
	    g_free(s);

	    s = g_strdup_printf("%i", minor);
	    gtk_entry_set_text(GTK_ENTRY(p->minor_entry), s);
	    g_free(s);
	}

	p->freeze_count--;
}

/*
 *	Apply VFS object callback.
 */
gboolean EDVPropertiesDeviceNodePageApplyVFSCB(
	EDVPropertiesDialog *d,
	const EDVObjectType type,
	GList *properties_list,
	gpointer data
)
{
	gboolean status = FALSE;
	gint		major,
			minor,
			cur_major,
			cur_minor;
	const gchar *path;
	GtkWidget *toplevel = d->toplevel;
	EDVCore *core = d->core;
	EDVPropertiesDeviceNodePage *p = EDV_PROPERTIES_DEVICE_NODE_PAGE(data);

	p->freeze_count++;

        /* Get the values that are needed to refer to the object that
         * we want to apply this page's values to
         */
	path = edv_properties_list_get_s(
	    properties_list,
	    EDV_PROP_NAME_PATH
	);
	if(STRISEMPTY(path))
	{
	    p->freeze_count--;
	    return(status);
	}

	/* Get the new major & minor numbers to be set */
	major = ATOI(gtk_entry_get_text(GTK_ENTRY(p->major_entry))),
	minor = ATOI(gtk_entry_get_text(GTK_ENTRY(p->minor_entry)));

	/* Get the current major & minor numbers */
	edv_device_numbers_parse(
	    edv_properties_list_get_i(
		properties_list,
		"device_type"
	    ),
	    &cur_major, &cur_minor
	);

	/* Do we need to change the values? */
	if((major != cur_major) || (minor != cur_minor))
	{
	    /* Set the device number by recreating the device node */
	    const gint device_type = edv_device_numbers_format(
		major, minor
	    );

	    /* Remove the existing device */
	    if(edv_unlink(path))
	    {
		const gint error_code = (gint)errno;
		gchar *msg = g_strdup_printf(
"Unable to change the device numbers of:\n\
\n\
    %s\n\
\n\
%s.",
		    path,
		    g_strerror(error_code)
		);
		edv_play_sound_error(core);
		edv_message_error(
		    "Change Device Numbers Error",
		    msg,
		    NULL,
		    toplevel
		);
		g_free(msg);
	    }
	    /* Recreate the device */
	    else if(mknod(
		(const char *)path,
		edv_edv_permissions_to_stat_mode((EDVPermissionFlags)edv_properties_list_get_i(
		    properties_list,
		    "permissions"
		)),
		(dev_t)device_type
	    ))
	    {
		const gint error_code = (gint)errno;
		gchar *msg = g_strdup_printf(
"Unable to change the device numbers of:\n\
\n\
    %s\n\
\n\
%s.",
		    path,
		    g_strerror(error_code)
		);
		edv_play_sound_error(core);
		edv_message_error(
		    "Change Device Numbers Error",
		    msg,
		    NULL,
		    toplevel
		);
		g_free(msg);
	    }
	    else
	    {
		/* Restore the ownership */
		(void)edv_lchown(
		    path,
		    edv_properties_list_get_i(
			properties_list,
			"owner_id"
		    ),
		    edv_properties_list_get_i(
			properties_list,
			"group_id"
		    )
		);

		/* Restore the time stamps */
		(void)edv_utime(
		    path,
		    edv_properties_list_get_ul(
			properties_list,
			"access_time"
		    ),
		    edv_properties_list_get_ul(
			properties_list,
			"modify_time"
		    )
		);

		status = TRUE;
	    }
	}

	p->freeze_count--;

	return(status);
}

/*
 *      Apply recycled object callback.
 */
gboolean EDVPropertiesDeviceNodePageApplyRecBinCB(
	EDVPropertiesDialog *d,
	const EDVObjectType type,
	GList *properties_list,
	gpointer data
)
{
	EDVPropertiesDeviceNodePage *p = EDV_PROPERTIES_DEVICE_NODE_PAGE(data);

	p->freeze_count++;

	p->freeze_count--;

	return(FALSE);
}


/*
 *      Destroy callback.
 */
void EDVPropertiesDeviceNodePageDestroyCB(gpointer data)
{
	EDVPropertiesDeviceNodePage *p = EDV_PROPERTIES_DEVICE_NODE_PAGE(data);
	g_free(p);
}
