/*
		       Endeavour Recycle Bin
 */

#ifndef RECBIN_H
#define RECBIN_H

#include <gtk/gtk.h>
#include "cfg.h"
#include "edv_types.h"
#include "tool_bar.h"
#include "guiutils.h"


typedef struct _edv_recbin_struct	edv_recbin_struct;
#define EDV_RECBIN(p)			((edv_recbin_struct *)(p))

typedef struct _edv_recbin_opid_struct	edv_recbin_opid_struct;
#define EDV_RECBIN_OPID(p)		((edv_recbin_opid_struct *)(p))


/*
 *	Flags:
 */
typedef enum {
	EDV_RECBIN_MAPPED		= (1 << 0),
	EDV_RECBIN_REALIZED		= (1 << 1),
	EDV_RECBIN_PROCESSING		= (1 << 2),
	EDV_RECBIN_TOOL_BAR_MAPPED	= (1 << 4),
	EDV_RECBIN_FIND_BAR_MAPPED	= (1 << 5),
	EDV_RECBIN_STATUS_BAR_MAPPED	= (1 << 6)
} edv_recbin_flags;


/*
 *	Recycle Bin Contents List Column Types:
 */
typedef enum {
	EDV_RECBIN_COLUMN_TYPE_INDEX			= 0,
	EDV_RECBIN_COLUMN_TYPE_NAME			= 1,
	EDV_RECBIN_COLUMN_TYPE_SIZE			= 2,
	EDV_RECBIN_COLUMN_TYPE_STORAGE_SIZE		= 3,
	EDV_RECBIN_COLUMN_TYPE_TYPE			= 4,
	EDV_RECBIN_COLUMN_TYPE_PERMISSIONS		= 5,
	EDV_RECBIN_COLUMN_TYPE_OWNER			= 6,
	EDV_RECBIN_COLUMN_TYPE_GROUP			= 7,
	EDV_RECBIN_COLUMN_TYPE_DATE_ACCESS		= 8,
	EDV_RECBIN_COLUMN_TYPE_DATE_MODIFIED		= 9,
	EDV_RECBIN_COLUMN_TYPE_DATE_CHANGED		= 10,
	EDV_RECBIN_COLUMN_TYPE_DATE_DELETED		= 11,
	EDV_RECBIN_COLUMN_TYPE_LINKED_TO		= 12,
	EDV_RECBIN_COLUMN_TYPE_ORIGINAL_LOCATION	= 13,
	EDV_RECBIN_COLUMN_TYPE_CAPACITY_USED		= 14
} edv_recbin_column_type;
/* Total number of column types */
#define EDV_RECBIN_CONTENTS_CLIST_COLUMNS_MAX		15


/*
 *	Recycle Bin Operations:
 */
typedef enum {

	EDV_RECBIN_OP_NONE			= 0,
	EDV_RECBIN_OP_SEPARATOR			= 100,

	EDV_RECBIN_OP_CLOSE			= 200,
	EDV_RECBIN_OP_EXIT			= 201,

	EDV_RECBIN_OP_SYNC_DISKS		= 301,
	EDV_RECBIN_OP_HISTORY			= 302,
	EDV_RECBIN_OP_RUN			= 303,
	EDV_RECBIN_OP_RUN_TERMINAL		= 304,
	EDV_RECBIN_OP_WRITE_PROTECT		= 305,
	EDV_RECBIN_OP_DELETE_METHOD_RECYCLE	= 308,
	EDV_RECBIN_OP_DELETE_METHOD_PURGE	= 309,

	EDV_RECBIN_OP_SYNC_RECYCLE_BIN		= 350,

	EDV_RECBIN_OP_RECOVER			= 700,
	EDV_RECBIN_OP_PURGE			= 701,
	EDV_RECBIN_OP_PURGE_ALL			= 702,
	EDV_RECBIN_OP_RENAME			= 703,
	EDV_RECBIN_OP_CHMOD			= 704,
	EDV_RECBIN_OP_CHOWN			= 705,
	EDV_RECBIN_OP_CHTIME			= 706,
	EDV_RECBIN_OP_PROPERTIES		= 707,

	EDV_RECBIN_OP_SELECT_ALL		= 800,
	EDV_RECBIN_OP_UNSELECT_ALL		= 801,
	EDV_RECBIN_OP_INVERT_SELECTION		= 802,
	EDV_RECBIN_OP_FIND			= 850,

	EDV_RECBIN_OP_REFRESH			= 1000,
	EDV_RECBIN_OP_REFRESH_ALL		= 1001,

	EDV_RECBIN_OP_SHOW_TOOL_BAR		= 1300,
	EDV_RECBIN_OP_SHOW_FIND_BAR		= 1303,
	EDV_RECBIN_OP_SHOW_STATUS_BAR		= 1304,

	EDV_RECBIN_OP_CONTENTS_LIST_FILTER	= 1400,
	EDV_RECBIN_OP_CONTENTS_LIST_AUTO_RESIZE_COLUMNS = 1410,

	EDV_RECBIN_OP_MIME_TYPES		= 1700,

	EDV_RECBIN_OP_NEW_BROWSER		= 1800,
	EDV_RECBIN_OP_NEW_IMBR			= 1801,
	EDV_RECBIN_OP_NEW_ARCHIVER		= 1802,

	EDV_RECBIN_OP_OPTIONS			= 2000,
	EDV_RECBIN_OP_CUSTOMIZE			= 2001,

	EDV_RECBIN_OP_HELP_CONTENTS		= 3000,
	EDV_RECBIN_OP_HELP_FILE_BROWSER		= 3010,
	EDV_RECBIN_OP_HELP_IMAGE_BROWSER	= 3011,
	EDV_RECBIN_OP_HELP_ARCHIVER		= 3012,
	EDV_RECBIN_OP_HELP_RECYCLE_BIN		= 3030,
	EDV_RECBIN_OP_HELP_KEYS_LIST		= 3050,
	EDV_RECBIN_OP_HELP_COMMON_OPERATIONS	= 3053,
	EDV_RECBIN_OP_HELP_ABOUT		= 3100

} edv_recbin_op;


/*
 *	Recycle Bin Operation ID:
 */
struct _edv_recbin_opid_struct {

	edv_recbin_op	op;
	EDVOpIDFlags	flags;
	ToolBarItemType	tb_item_type;
	edv_recbin_struct	*recbin;

	/* Operation callback */
	void		(*func_cb)(
		ToolBarItem *,			/* Tool Bar Item */
		gint,				/* ID */
		gpointer			/* OPID */
	);

	/* Enter and leave notify callbacks */
	void		(*enter_func_cb)(
		ToolBarItem *,			/* Tool Bar Item */
		gint,				/* ID */
		gpointer			/* OPID */
	);
	void		(*leave_func_cb)(
		ToolBarItem *,			/* Tool Bar Item */
		gint,				/* ID */
		gpointer			/* OPID */
	);

	gchar		*button_name;		/* Name on a button */
	gchar		*menu_name;		/* Name on a menu item */
	gchar		*tooltip;		/* Tool tip or description (one line) */

	guint		accel_key,		/* Accelerator key */
			accel_mods;		/* Accelerator key modifiers */

	guint8		**button_icon_data,	/* Shared */
			**menu_icon_data;	/* Shared */

};

/*
 *	Recycle Bin:
 */
struct _edv_recbin_struct {

	GtkWidget	*toplevel;
	EDVCore *core;
	edv_recbin_flags	flags;
	gint		freeze_count,
			busy_count,
			stop_count;

	GList		*opid_list;


	GtkWidget	*main_vbox,
			*menu_bar,

			*menu_bar_handle,
			*tool_bar_handle,
			*find_bar_handle,

			*contents_clist;	/* Contents GtkCList */

	/* GtkMenus and GtkMenuItems */
	GtkWidget	*recover_mi,
			*purge_mi,
			*purge_all_mi,
			*sync_recycle_bin_mi,
			*sync_disks_mi,
			*history_mi,
			*run_mi,
			*close_mi,

			*edit_recover_mi,
			*edit_purge_mi,
			*edit_purge_all_mi,
			*rename_mi,
			*chmod_mi,
			*chown_mi,
			*chtime_mi,
			*select_all_mi,
			*unselect_all_mi,
			*invert_selection_mi,
			*find_mi,
			*properties_mi,

			*refresh_mi,
			*refresh_all_mi,
			*show_tool_bar_micheck,
			*show_find_bar_micheck,
			*show_status_bar_micheck,
			*contents_list_filter_mi,
			*contents_list_auto_resize_columns_micheck,
			*mime_types_mi,

			*write_protect_mi,
			*delete_method_recycle_micheck,
			*delete_method_purge_micheck,
			*options_mi,
			*customize_mi,

			*window_new_browser_mi,
			*window_new_imbr_mi,
			*window_new_archiver_mi,

			*help_contents_mi,
			*help_recycle_bin_mi,
			*help_keys_list_mi,
			*help_common_operations_mi,
			*help_about_mi;

	/* Tool Bar */
	GtkWidget	*tool_bar;

	/* Find Bar */
	GtkWidget	*find_bar;

	/* Status Bar */
	GtkWidget	*status_bar;

	/* Contents GtkCList Right-Click GtkMenu */
	GtkWidget       *contents_clist_menu,
			*contents_clist_recover_mi,
			*contents_clist_purge_mi,
			*contents_clist_purge_all_mi,
			*contents_clist_rename_mi,
			*contents_clist_chmod_mi,
			*contents_clist_chown_mi,
			*contents_clist_chtime_mi,
			*contents_clist_properties_mi;

	gchar		*contents_list_filter;

	/* Last selected row on the clist (can be -1 for none)
	 *
	 * Note that this does not reflect other selected rows since
	 * multiple selections are allowed
	 */
	gint		contents_clist_selected_row;


	/* Last known items in recycle bin, this is used as a quick access
	 * value to check if the recycle bin has contents or not
	 *
	 * Can be -1 to indicate undefined
	 */
	gint		last_nrecycle_bin_items;

	/* Last known write protect state (can be -1 to indicate
	 * undefined)
	 */
	gint		last_write_protect_state;

};

#define EDV_RECBIN_IS_PROCESSING(p)	(((p) != NULL) ? \
 ((p)->flags & EDV_RECBIN_PROCESSING) : FALSE)
#define EDV_RECBIN_IS_FROZEN(p)		(((p) != NULL) ? \
 ((p)->freeze_count > 0) : FALSE)
#define EDV_RECBIN_IS_BUSY(p)		(((p) != NULL) ? \
 ((p)->busy_count > 0) : FALSE)


/* Utilities */
extern edv_recbin_opid_struct *EDVRecBinMatchOPID(
	edv_recbin_struct *recbin,
	const edv_recbin_op op
);

/* Operation IDs */
extern edv_recbin_opid_struct *EDVRecBinOPIDNew(void);
extern void EDVRecBinOPIDDelete(edv_recbin_opid_struct *opid);

/* Title */
extern void EDVRecBinSetTitle(edv_recbin_struct *recbin);

/* Get selected object & paths */
extern GList *EDVRecBinGetSelectedObjectsList(
	edv_recbin_struct *recbin, 
	const gboolean copy
);

/* Regenerate Style, Menus, and Tool Bar */
extern void EDVRecBinAccelkeysRegenerate(edv_recbin_struct *recbin);
extern void EDVRecBinToolBarRegenerate(edv_recbin_struct *recbin);
extern void EDVRecBinContentsMenuRegenerate(edv_recbin_struct *recbin);

/* Sync Configuration */
extern void EDVRecBinSyncConfiguration(edv_recbin_struct *recbin);

/* Operation IDs List */
extern GList *EDVRecBinOPIDsListNew(edv_recbin_struct *recbin);
extern void EDVRecBinOPIDsListDelete(GList *opids_list);

/* Accel Keys */
extern CfgAccelkeyList *EDVRecBinAccelkeysListNewDefault(void);

/* Recycle Bin */
extern edv_recbin_struct *EDVRecBinNew(
	EDVCore *core,
	const GdkGeometryFlags geometry_flags,
	const GdkRectangle *geometry
);
extern void EDVRecBinSyncData(edv_recbin_struct *recbin);
extern void EDVRecBinUpdateMenus(edv_recbin_struct *recbin);
extern void EDVRecBinSetBusy(edv_recbin_struct *recbin, const gboolean busy);
extern gboolean EDVRecBinIsMapped(edv_recbin_struct *recbin);
extern void EDVRecBinMap(edv_recbin_struct *recbin);
extern void EDVRecBinUnmap(edv_recbin_struct *recbin);
extern void EDVRecBinDelete(edv_recbin_struct *recbin);


#endif	/* RECBIN_H */
